/*
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#include <gtk/gtk.h>
#include "select.h"
#include "throw.h"
#include "position.h"
#include "set.h"
#include "game.h"
#include "free.h"
#include "draw.h"
#include "def.h"

/* game_play_ia: play automagically. That's 
 * what we call Artificial Intelligence.
 * Indeed that's not intelligent at all.
 */ 
void game_play_ia(struct _prog *prog_data)
{
	while(prog_data->player != 2 
	&& prog_data->players[prog_data->player] != NULL)
	{
		game_turn(prog_data, NULL);
	}
}

/* game_middle_position: set the position (x,y)
 * of each card in the list represented at the 
 * center of the game area (the table).
 */ 
void game_middle_position(struct _prog *prog_data)
{
	GList *ptr;
	
	struct _card *data;
	
	int player = prog_data->hashand;
	
	if(player != -1)
	{
		for(ptr = g_list_last(prog_data->waiting); ptr; ptr = ptr->prev)
		{
			data = ptr->data;
			
			game_position_middle(data, prog_data, player);
			
			select_player_next(&player);

		}
	}
}

/* game_hand_position: set the position (x,y)
 * of the cards in each player hand.
 */ 
void game_hand_position(struct _prog *prog_data, int player)
{
	int x, y, w, h;
	enum eposition ep;
	
	int coef, start, max;
	
	gdk_drawable_get_size(prog_data->back, 
			&w, &h);
	
	if(player == -1)
	{
		start = 0;
		max = 4;
	}
	else
	{
		start = player;
		max = player+1;
	}
	
	for(player = start; player < max; player++)
	{
	
		switch(player)
		{
			case 0:
			y = 10;
			coef = 1;
			x = (prog_data->area->allocation.width - w) / 3;
			ep = EP_HORIZONTAL;
			break;
			
			case 1:
			x = prog_data->area->allocation.width - w -10;
			coef = 1;
			y = (prog_data->area->allocation.height - h) / 2;
			ep = EP_VERTICAL;
			break;
			
			case 2:
			y = prog_data->area->allocation.height - h -10;
			coef = -1;
			x = (prog_data->area->allocation.width - w)*2 / 3;
			ep = EP_HORIZONTAL;
			break;
			
			case 3:
			x = 10;
			coef = -1;
			y = (prog_data->area->allocation.height - h)*2 / 3;
			ep = EP_VERTICAL;
			break;
		}
		
		position_list(prog_data->players[player], 
			x, y, ep, coef, prog_data);
		
	}
}

/* game_send: distributing cards function.
 * This will send cards from one list to 
 * players list. Each player will take 
 * [the value of mod] cards.
 * 
 * int mod: how many cards per player.
 * 
 * int max: how many cards we want to send 
 * (total)
 * 
 * GList *ptr: the list where we take the 
 * cards.
 */ 
GList* game_send(struct _prog *prog_data, int mod, int max,
	GList *ptr)
{
	int player = prog_data->distributer;
	int i, modifier;
	GList *last;
	
	select_player_next(&player);
	
	for(modifier = 0, i = 1; i <= max; i++)
	{
		if(ptr != NULL)
		{
			throw_to(&prog_data->players[player],
				(struct _card*) ptr->data);
			
			if(player == 2)
			{
				((struct _card*) ptr->data)->draw_face = TRUE;
			}
			
			/* at the second distribution phase, 
			 * we need to send three cards to each player 
			 * except the one that choose the trump family.
			 * This one will take only two cards.
			 * That's why we do i++ here, so we won't be 
			 * distributing [the value of mod] cards to him, 
			 * but instead [[the value of mod]-1] cards. */
			if(player == prog_data->master && modifier == 0)
			{
				i++;
				modifier = 1;
			}
				
			if(i % mod == 0)
			{
				select_player_next(&player);
			}	
			
			
			last = ptr;
			ptr = ptr->next;

			prog_data->pile = g_list_remove(prog_data->pile, 
				last->data);
			
		}
		else i = max+1;
	}
	
	return(ptr);
}

/* game_distribute_phase1 : send five cards to 
 * each player. (Three first, and two)
 */ 
void game_distribute_phase1(struct _prog *prog_data)
{
	if(prog_data->players && prog_data->pile)
	{
		prog_data->pile = game_send(prog_data, 3, 12,
			prog_data->pile);
		
		prog_data->pile = game_send(prog_data, 2, 8,
			prog_data->pile);
		
		game_hand_position(prog_data, -1);
	}
}

/* game_distribute_phase2 : send remaining card in 
 * the pile to the players. The card which was 
 * proposed is to be sent to the master player.
 */ 
void game_distribute_phase2(struct _prog *prog_data)
{
	if(prog_data->players && prog_data->pile)
	{
		prog_data->pile = game_send(prog_data, 3, 12,
			prog_data->pile);
		
		if(prog_data->waiting)
		{
			struct _card *data_ptr = (struct _card*) prog_data->waiting->data;
			
			if(prog_data->master != 2)
				data_ptr->draw_face = FALSE;
			
			data_ptr->moveable = FALSE;
		
			throw_to(&prog_data->players[prog_data->master],
				data_ptr);
		
			prog_data->waiting = g_list_remove(prog_data->waiting, 
				data_ptr);
		}
		
		game_hand_position(prog_data, -1);
		
		set_moveable(prog_data->players[2], TRUE);
	}
}

/* game_finish_phase1 : just take one card from 
 * the pile and put it on the top of the pile, 
 * with the value displayed to the user. 
 * This is the proposed trump card.
 */ 
void game_finish_phase1(struct _prog *prog_data)
{
	GList *pile = prog_data->pile;
	
	if(pile != NULL)
	{
		struct _card *data_ptr = (struct _card*) pile->data;
		
		throw_to(&prog_data->waiting,
			data_ptr);
		
		data_ptr->draw_face = TRUE;
		data_ptr->moveable = TRUE;
		
		prog_data->pile = g_list_remove(prog_data->pile, 
				data_ptr);
	}
}

/* game_drop_middle: there should be four 
 * cards on the table when calling this 
 * function. It means we need to know 
 * what team wins the turn, and send 
 * all of those cards to the winner team 
 * list.

 * Note: indeed the winner player of
 * this turn has already been determined, 
 * it's the 'menext' value in struct _prog
 * That's why we don't need to find the 
 * winner team here, we know it already
 * (it's [menext MODULO 2])
 */ 
void game_drop_middle(struct _prog *prog_data)
{
	struct _card *data_ptr;
	GList *ptr;
	int winner = -1;
	
	if(prog_data->menext >= 0)
	{
		winner = prog_data->menext % 2;
		
		
		fprintf(prog_data->output, "L'équipe %d", winner);
		if(winner == 0)
			fprintf(prog_data->output, "(0 = haut, 2 = vous)");
		else
			fprintf(prog_data->output, "(1 = droite, 3 = gauche)");
		fprintf(prog_data->output, " empoche le pli\n");
		
		
		ptr = prog_data->waiting;
		
		while( ptr )
		{
			data_ptr = (struct _card *)ptr->data;
			
			data_ptr->draw = FALSE;
			
			throw_to(&prog_data->teams[winner], data_ptr);
			
			ptr = ptr->next;
			
		}
		
		g_list_free(prog_data->waiting);
		
		prog_data->waiting = NULL;
	}
	else
	{
		fprintf(prog_data->output, "Erreur: impossible de déterminer l'équipe qui remporte le pli\n");
	}
}

/* game_position_middle : set the position (x,y) 
 * of a card on the table. Each card is 
 * positionned in relation to the player 
 * who dropped it.
 */ 
void game_position_middle(struct _card *card, 
	struct _prog *prog_data,
	int player)
{
	int midx, midy;
	
	midx = (prog_data->area->allocation.width - card->dim.w)/2;
	midy = (prog_data->area->allocation.height - card->dim.h)/2;
	
	switch(player)
	{
		case 0:
		midy -= 15;
		break;
		
		case 1:
		midx += 15;
		break;
		
		case 2:
		midy += 15;
		break;
		
		case 3:
		midx -= 15;
		break;
	}
	
	if(card->dim.x != midx || card->dim.y != midy)
	{
		if(prog_data->animation)
			position_move_card(card, midx, midy, prog_data);
		

		card->dim.x = midx;
		card->dim.y = midy;
		draw_container(prog_data);
	}
}

/* game_print_family: that's in french atm,
 * feel free to translate.
 */ 
void game_print_family(FILE *stream, int family)
{
	switch(family)
	{
		case 0:
		/* diamond */
		fprintf(stream, "carreau");
		break;
		case 1:
		/* heart */
		fprintf(stream, "coeur");
		break;
		case 2:
		/* club */
		fprintf(stream, "pique");
		break;
		case 3:
		/* spade */
		fprintf(stream, "trefle");
		break;
	}

}

/* game_print_card: print the card as 
 * a nominative value.
 * Example: ace of diamond
 * Feel free to translate.
 */ 
void game_print_card(FILE *stream, struct _card *card)
{
	switch(card->num)
	{
		case 0:
		/* ace */
		fprintf(stream, "as");
		break;
		
		case 6:
		case 7:
		case 8:
		case 9:
		fprintf(stream, "%d", card->num+1);
		break;
		
		case 10:
		/* jack */
		fprintf(stream, "valet");
		break;
		
		case 11:
		/* queen */
		fprintf(stream, "dame");
		break;
		
		case 12:
		/* king */
		fprintf(stream, "roi");
		break;
	}
	
	/* of */
	fprintf(stream, " de ");
	
	game_print_family(stream, card->family);
}

/* game_get_family_best: what is the most 
 * powerfull card we get on a GList in 
 * a particular family only.
 */ 
int game_get_family_best(int family, GList *lst)
{
	int nb = -1, i, best = 0;
	
	struct _card *data;
	
	for(i=0; lst; i++)
	{
		
		data = (struct _card*)lst->data;
		
		if(data->family == family)
		{
			if(data->points > best)
			{
				best = data->points;
				nb = i;
			}
		}
		/*else
		if(data->family == majorfamily && best == 0)
		{
			best = data->points;
			family = data->family;
			nb = i;
		}*/

		lst = lst->next;
	}
	
	return(nb);
}

/* game_get_best: what is the most 
 * powerful card on a list, related 
 * to a family, but not only. 
 * If we found a trump card, it will 
 * probably be better than something 
 * else. That's what majorfamily is 
 * all about.
 * 
 * Note: if the family value is -1, 
 * then the first card family is 
 * taken as a reference to find more 
 * powerfull cards. Don't forget 
 * that if majorfamily is set (other 
 * than -1), and if it's different 
 * than the value of family, any 
 * card of the majorfamily is better 
 * then the first card.
 * 
 * That function is heavily used to 
 * find if a card dropped on the 
 * table is better than others.
 * 
 * We just find what is the best 
 * card on the table. Then we check 
 * this card against the one to be 
 * dropped.
 */ 
struct _card* game_get_best(int *best, int *family, 
	int majorfamily, GList *lst)
{
	struct _card *data, *first = NULL;
	
	lst = g_list_last(lst);
	
	while(lst)
	{
		
		data = (struct _card*)lst->data;
		
		if(*family == -1)
		{
			first = data;
			*best = data->points;
			*family = data->family;
		}
		else
		if(data->family == *family)
		{
			if(data->points > *best)
				*best = data->points;
		}
		else
		if(data->family == majorfamily)
		{
			*best = data->points;
			*family = data->family;
		}
			
		lst = lst->prev;
	}
	
	return(first);
}

/* game_check_cardisbetter: check if a card 
 * is more powerfull than the most powerfull
 * card on a GList
 */ 
gboolean game_check_cardisbetter(struct _card *card, GList *lst,
	int majorfamily)
{
	int best = 0, family = -1;
	gboolean better = FALSE;
	
	struct _card *first;
	
	first = game_get_best(&best, &family, majorfamily, lst);
	
	if(best > 0)
	{
		if(card->family == family)
		{
			if(card->points > best)
				better = TRUE;
		}
		else if(card->family == majorfamily)
		{
			better = TRUE;
		}
	}
	
	return(better);
}

/* game_ia_select_card: what the computer should 
 * play. This is where every decision should be 
 * made. 
 * 
 * 
 * Note: the rules of the game should be excluded 
 * from here and put on another function, because 
 * the human player is also affected by the rules.
 * --Need to be Fixed--
 */ 
int game_ia_select_card(struct _prog *prog_data)
{
	int nb=0, best = 0, family = -1;
	
	struct _card *first = NULL;
	
	if(prog_data->waiting != NULL)
	{
		first = game_get_best(&best, &family, prog_data->family, 
			prog_data->waiting);
	}
	
	if(first != NULL)
	{
		nb = game_get_family_best(first->family, 
			prog_data->players[prog_data->player]);
		
		if(nb < 0)
		{
			/*fprintf(prog_data->output, 
				"Le joueur %d n'a pas de ", prog_data->player);
			game_print_family(prog_data->output, first->family);
			*/
			if(prog_data->family != first->family)
			{
				nb = game_get_family_best(prog_data->family, 
				prog_data->players[prog_data->player]);
				/*if(nb > 0)
					fprintf(prog_data->output, " mais il a de l'atout");
				else
					fprintf(prog_data->output, " et n'a pas d'atout");*/
			}
			/*fprintf(prog_data->output, "\n");*/
			
		}
		
		if(nb < 0)
			nb = 0;
	}
	
	return(nb);
}

/* game_start: makes the main list  
 * of cards, called the pile.
 * That's where we take care of 
 * preparing that pile and the 
 * status of each card before 
 * starting a new round. 
 */ 
void game_start(struct _prog *prog_data)
{
	throw_card_list( prog_data->all );
	
	throw_cards(prog_data);
	
	prog_data->state = READY_PILE;
	
	gtk_widget_set_sensitive(prog_data->button, TRUE);
}

/* game_totalcards: calculate the 
 * total value of points in a GList. 
 */ 
int game_totalcards(GList *lst)
{
	int total = 0;
	
	struct _card *data;
	
	while(lst)
	{
		data = lst->data;
		
		 /* Warning: the points value of cards 
		 * are not set to the real value 
		 * of the card in the game, but at 
		 * a more powerfull value.
		 * The real value is [value/10] */
		total += data->points/10;
		
		lst = lst->next;
	}
	
	return(total);
}

/* game_restart: reset state and int values 
 * of the game before each new rounds.
 */ 
void game_restart(struct _prog *prog_data)
{
	prog_data->state = LOADING;
	prog_data->player = -1;
	prog_data->master = -1;
	prog_data->family = -1;
	
	prog_data->hashand = -1;
	prog_data->menext = -1;
	
}

/* game_turn: so complicated. 
 * --Need to be simplified--
 */ 
void game_turn(struct _prog *prog_data, struct _card *card)
{
	static int turn = 8;
	int cardnb;
	
	if(prog_data->player == prog_data->hashand)
	{
		if(prog_data->waiting != NULL)
		{
			game_drop_middle(prog_data);
			
			if(prog_data->menext >= 0)
			{
				prog_data->hashand = prog_data->menext;
				prog_data->player = prog_data->menext;
			}
		}
	}
	
	if(turn > 0)
	{
		if(card == NULL && prog_data->player != 2)
		{
			/*if(prog_data->player == 2)
				cardnb = gtk_spin_button_get_value_as_int(
					GTK_SPIN_BUTTON(prog_data->allwidgets[4]))-1;
			else*/
				cardnb = game_ia_select_card(prog_data);
			
			card = throw_card_x(prog_data->players[prog_data->player], 
				cardnb);
		}
		
		

		if(card != NULL)
		{
			
			fprintf(prog_data->output, 
				"Joueur %d joue la carte ", 
				prog_data->player);
			game_print_card(prog_data->output, card);
			fprintf(prog_data->output, "\n");
			
			
			if(prog_data->waiting != NULL 
			&& game_check_cardisbetter(card, prog_data->waiting, 
				prog_data->family))
			{
				fprintf(prog_data->output, "Joueur %d prend la main\n", prog_data->player);
				prog_data->menext = prog_data->player;
			}
			
			throw_to(&prog_data->waiting, card);
			
			card->draw_face = TRUE;
			
			prog_data->players[prog_data->player] = g_list_remove(
				prog_data->players[prog_data->player], 
				(gconstpointer) card);
			
			game_hand_position(prog_data, prog_data->player);
		
		
			
			game_position_middle(card, prog_data, prog_data->player);
			
			select_player_next(&prog_data->player);
			
			if(prog_data->player == 2)
				gtk_widget_set_sensitive(prog_data->allwidgets[4], TRUE);
			else
				gtk_widget_set_sensitive(prog_data->allwidgets[4], FALSE);
			
			if(prog_data->player == prog_data->hashand)
			{
				if(prog_data->menext == 2)
					gtk_widget_set_sensitive(prog_data->allwidgets[4], TRUE);
				else
					gtk_widget_set_sensitive(prog_data->allwidgets[4], FALSE);
					
				turn--;
				if(turn > 0)
					gtk_spin_button_set_range(
						GTK_SPIN_BUTTON(prog_data->allwidgets[4]),
						1, turn);
				/*else
				{
					gtk_widget_set_sensitive(prog_data->allwidgets[2], FALSE);
					gtk_widget_set_sensitive(prog_data->allwidgets[4], FALSE);
					turn = 8;
				}*/
			}
		}
	}
	else
	{
		prog_data->state = ENDGAME;
		/*gtk_widget_set_sensitive(prog_data->allwidgets[2], FALSE);*/
		gtk_widget_set_sensitive(prog_data->allwidgets[4], FALSE);
		
		fprintf(prog_data->output, 
			"Points des cartes: [0,2] %d (vous) contre %d [1,3]\n",
			game_totalcards(prog_data->teams[0]),
			game_totalcards(prog_data->teams[1]));
		
		if(prog_data->allwidgets && prog_data->allwidgets[5])
		{
			if(gtk_image_get_pixbuf(GTK_IMAGE(prog_data->allwidgets[5])))
				gtk_image_clear(GTK_IMAGE(prog_data->allwidgets[5]) );
		}
	
		free_clear(prog_data);
		game_restart(prog_data);
		turn = 8;
		game_start(prog_data);
		/*
		gtk_spin_button_set_range(
			GTK_SPIN_BUTTON(prog_data->allwidgets[4]),
			1, turn);*/
	}
}
